import unittest
from xevents import xEvents


class TestFilter(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    def test_filter_skips_non_matching(self):
        self.bus.subscribe(
            "order",
            lambda d: self.received.append(d),
            filter_fn=lambda d: d.get("status") == "paid",
        )
        self.bus.post("order", {"status": "pending", "id": 1})
        self.bus.post("order", {"status": "paid", "id": 2})
        self.bus.post("order", {"status": "refunded", "id": 3})
        self.assertEqual(len(self.received), 1)
        self.assertEqual(self.received[0]["id"], 2)


if __name__ == "__main__":
    unittest.main()



import unittest
from xevents import xEvents


class TestReplay(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    def test_replay_delivers_past_events(self):
        self.bus.post("update", {"v": 1})
        self.bus.post("update", {"v": 2})
        self.bus.post("other", {"v": 3})
        count = self.bus.replay("update", lambda d: self.received.append(d))
        self.assertEqual(count, 2)
        self.assertEqual(len(self.received), 2)

    def test_replay_with_limit(self):
        for i in range(5):
            self.bus.post("x", {"i": i})
        count = self.bus.replay("x", lambda d: self.received.append(d), limit=2)
        self.assertEqual(count, 2)
        self.assertEqual(self.received[0]["i"], 3)

    def test_replay_channel(self):
        self.bus.post("orders:created", {"id": 1})
        self.bus.post("orders:updated", {"id": 2})
        self.bus.post("users:created", {"id": 3})
        count = self.bus.replay_channel("orders", lambda d: self.received.append(d))
        self.assertEqual(count, 2)


if __name__ == "__main__":
    unittest.main()



import unittest
from xevents import xEvents


class TestReset(unittest.TestCase):
    def test_reset_clears_everything(self):
        bus = xEvents()
        bus.subscribe("a", lambda d: None)
        bus.channel("ch")
        bus.post("a", {"v": 1})
        bus.on_any("*", lambda e, d: None)
        bus.reset()
        self.assertEqual(bus.events(), [])
        self.assertEqual(bus.channels(), [])
        self.assertEqual(bus.history(), [])


class TestListenerCount(unittest.TestCase):
    def test_count(self):
        bus = xEvents()
        bus.subscribe("a", lambda d: None)
        bus.subscribe("a", lambda d: None)
        self.assertEqual(bus.listener_count("a"), 2)
        self.assertEqual(bus.listener_count("nonexistent"), 0)


class TestErrorHandling(unittest.TestCase):
    def test_listener_error_does_not_break_others(self):
        bus = xEvents()
        results = []

        def bad_listener(d):
            raise RuntimeError("boom")

        def good_listener(d):
            results.append(d)

        bus.subscribe("test", bad_listener)
        bus.subscribe("test", good_listener)
        bus.post("test", {"a": 1})
        self.assertEqual(len(results), 1)


class TestReentrancy(unittest.TestCase):
    def test_listener_can_post_events(self):
        bus = xEvents()
        results = []

        def on_a(d):
            results.append(("a", d))
            if d.get("trigger_b"):
                bus.post("b", {"from": "a"})

        def on_b(d):
            results.append(("b", d))

        bus.subscribe("a", on_a)
        bus.subscribe("b", on_b)
        bus.post("a", {"trigger_b": True})
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0][0], "a")
        self.assertEqual(results[1][0], "b")


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass, field
from typing import Any, Optional
import time


@dataclass
class EventRecord:
    event: str
    data: dict[str, Any]
    timestamp: float = field(default_factory=time.time)
    channel: Optional[str] = None


@dataclass(frozen=True)
class _ListenerKey:
    event: str
    listener_id: int



import unittest
from xevents import xEvents, EventRecord


class TestHistory(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents(history_limit=50)

    def test_history_records(self):
        self.bus.post("a", {"v": 1})
        self.bus.post("b", {"v": 2})
        h = self.bus.history()
        self.assertEqual(len(h), 2)
        self.assertIsInstance(h[0], EventRecord)

    def test_history_filter_by_event(self):
        self.bus.post("a", {"v": 1})
        self.bus.post("b", {"v": 2})
        self.bus.post("a", {"v": 3})
        h = self.bus.history(event="a")
        self.assertEqual(len(h), 2)

    def test_history_filter_by_channel(self):
        self.bus.post("orders:created", {"id": 1})
        self.bus.post("users:created", {"id": 2})
        h = self.bus.history(channel="orders")
        self.assertEqual(len(h), 1)
        self.assertEqual(h[0].data["id"], 1)

    def test_history_limit(self):
        self.bus.post("a", {"v": 1})
        self.bus.post("a", {"v": 2})
        self.bus.post("a", {"v": 3})
        h = self.bus.history(event="a", limit=2)
        self.assertEqual(len(h), 2)
        self.assertEqual(h[0].data["v"], 2)

    def test_history_cap(self):
        bus = xEvents(history_limit=5)
        for i in range(10):
            bus.post("x", {"i": i})
        h = bus.history()
        self.assertEqual(len(h), 5)
        self.assertEqual(h[0].data["i"], 5)

    def test_clear_history(self):
        self.bus.post("a", {})
        self.bus.clear_history()
        self.assertEqual(len(self.bus.history()), 0)


if __name__ == "__main__":
    unittest.main()



import unittest
from xevents import xEvents


class TestSubscribeAndPost(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    def test_subscribe_and_post(self):
        self.bus.subscribe("test", lambda d: self.received.append(d))
        self.bus.post("test", {"msg": "hello"})
        self.assertEqual(len(self.received), 1)
        self.assertEqual(self.received[0]["msg"], "hello")

    def test_on_alias(self):
        self.bus.on("test", lambda d: self.received.append(d))
        self.bus.post("test", {"val": 1})
        self.assertEqual(len(self.received), 1)

    def test_unsubscribe(self):
        listener = lambda d: self.received.append(d)
        self.bus.subscribe("test", listener)
        self.bus.post("test", {"a": 1})
        self.bus.unsubscribe("test", listener)
        self.bus.post("test", {"a": 2})
        self.assertEqual(len(self.received), 1)

    def test_multiple_listeners(self):
        results_a = []
        results_b = []
        self.bus.subscribe("test", lambda d: results_a.append(d))
        self.bus.subscribe("test", lambda d: results_b.append(d))
        self.bus.post("test", {"x": 1})
        self.assertEqual(len(results_a), 1)
        self.assertEqual(len(results_b), 1)

    def test_disabled_events(self):
        self.bus.subscribe("test", lambda d: self.received.append(d))
        self.bus.enabled = False
        self.bus.post("test", {"a": 1})
        self.assertEqual(len(self.received), 0)

    def test_no_duplicate_subscribe(self):
        listener = lambda d: self.received.append(d)
        self.bus.subscribe("test", listener)
        self.bus.subscribe("test", listener)
        self.bus.post("test", {"a": 1})
        self.assertEqual(len(self.received), 1)

    def test_pre_registered_events(self):
        bus = xEvents(events=["start", "stop"])
        evts = bus.events()
        self.assertIn("start", evts)
        self.assertIn("stop", evts)


class TestOnceListener(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    def test_once_fires_once(self):
        self.bus.once("test", lambda d: self.received.append(d))
        self.bus.post("test", {"a": 1})
        self.bus.post("test", {"a": 2})
        self.assertEqual(len(self.received), 1)
        self.assertEqual(self.received[0]["a"], 1)

    def test_once_alias(self):
        self.bus.once("test", lambda d: self.received.append(d))
        self.bus.post("test", {"v": 10})
        self.bus.post("test", {"v": 20})
        self.assertEqual(len(self.received), 1)


if __name__ == "__main__":
    unittest.main()



from xevents.bus import xEvents
from xevents.channel import Channel
from xevents.models import EventRecord
from xevents.types import Event, EventData, Listener, Events
from xevents.constants import CHANNEL_SEP, WILDCARD

__all__ = [
    "xEvents",
    "Channel",
    "EventRecord",
    "Event",
    "EventData",
    "Listener",
    "Events",
    "CHANNEL_SEP",
    "WILDCARD",
]



import unittest
from xevents import xEvents


class TestWildcard(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    def test_wildcard_star(self):
        self.bus.on_any("orders:*", lambda ev, d: self.received.append((ev, d)))
        self.bus.post("orders:created", {"id": 1})
        self.bus.post("orders:deleted", {"id": 2})
        self.bus.post("users:created", {"id": 3})
        self.assertEqual(len(self.received), 2)
        self.assertEqual(self.received[0][0], "orders:created")
        self.assertEqual(self.received[1][0], "orders:deleted")

    def test_wildcard_all(self):
        self.bus.on_any("*", lambda ev, d: self.received.append(ev))
        self.bus.post("a", {})
        self.bus.post("b", {})
        self.assertEqual(len(self.received), 2)

    def test_off_any(self):
        listener = lambda ev, d: self.received.append(ev)
        self.bus.on_any("*", listener)
        self.bus.post("a", {})
        self.bus.off_any("*", listener)
        self.bus.post("b", {})
        self.assertEqual(len(self.received), 1)


if __name__ == "__main__":
    unittest.main()



import asyncio
import unittest
from xevents import xEvents


class TestAsyncListeners(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    async def test_async_listener(self):
        async def handler(d):
            self.received.append(d)

        self.bus.subscribe("test", handler)
        self.bus.post("test", {"v": 1})
        await asyncio.sleep(0.05)
        self.assertEqual(len(self.received), 1)
        self.assertEqual(self.received[0]["v"], 1)

    async def test_async_wildcard_listener(self):
        async def handler(ev, d):
            self.received.append((ev, d))

        self.bus.on_any("data:*", handler)
        self.bus.post("data:update", {"v": 2})
        await asyncio.sleep(0.05)
        self.assertEqual(len(self.received), 1)
        self.assertEqual(self.received[0][0], "data:update")


if __name__ == "__main__":
    unittest.main()



import unittest
from xevents import xEvents


class TestChannels(unittest.TestCase):
    def setUp(self):
        self.bus = xEvents()
        self.received = []

    def test_channel_subscribe_and_post(self):
        ch = self.bus.channel("orders")
        ch.subscribe("created", lambda d: self.received.append(d))
        ch.post("created", {"order_id": 1})
        self.assertEqual(len(self.received), 1)
        self.assertEqual(self.received[0]["order_id"], 1)

    def test_channel_isolation(self):
        orders = self.bus.channel("orders")
        users = self.bus.channel("users")
        order_data = []
        user_data = []
        orders.subscribe("created", lambda d: order_data.append(d))
        users.subscribe("created", lambda d: user_data.append(d))
        orders.post("created", {"order_id": 1})
        users.post("created", {"user_id": 5})
        self.assertEqual(len(order_data), 1)
        self.assertEqual(len(user_data), 1)
        self.assertEqual(order_data[0]["order_id"], 1)
        self.assertEqual(user_data[0]["user_id"], 5)

    def test_channel_on_alias(self):
        ch = self.bus.channel("tasks")
        ch.on("done", lambda d: self.received.append(d))
        ch.post("done", {"task": "a"})
        self.assertEqual(len(self.received), 1)

    def test_channel_once(self):
        ch = self.bus.channel("tasks")
        ch.once("done", lambda d: self.received.append(d))
        ch.post("done", {"task": "a"})
        ch.post("done", {"task": "b"})
        self.assertEqual(len(self.received), 1)

    def test_channel_unsubscribe(self):
        ch = self.bus.channel("tasks")
        listener = lambda d: self.received.append(d)
        ch.subscribe("done", listener)
        ch.post("done", {"a": 1})
        ch.unsubscribe("done", listener)
        ch.post("done", {"a": 2})
        self.assertEqual(len(self.received), 1)

    def test_channels_list(self):
        self.bus.channel("a")
        self.bus.channel("b")
        names = self.bus.channels()
        self.assertIn("a", names)
        self.assertIn("b", names)


if __name__ == "__main__":
    unittest.main()



from threading import RLock
from typing import Callable, List
from collections import defaultdict
from fnmatch import fnmatch
import asyncio
import logging

from xevents.channel import Channel
from xevents.constants import CHANNEL_SEP, DEFAULT_HISTORY_LIMIT
from xevents.models import EventRecord, _ListenerKey
from xevents.types import Event, EventData, Events, Listener

logger = logging.getLogger("xEvents")


class xEvents:
    def __init__(
        self,
        events: Events | None = None,
        debug: bool = False,
        history_limit: int = DEFAULT_HISTORY_LIMIT,
    ):
        self.__lock = RLock()
        self.__listeners: dict[Event, List[Listener]] = defaultdict(list)
        self.__once_listeners: set[_ListenerKey] = set()
        self.__wildcard_listeners: list[tuple[str, Listener]] = []
        self.__history: list[EventRecord] = []
        self.__history_limit = history_limit
        self.__channels: dict[str, Channel] = {}
        self.__filters: dict[_ListenerKey, Callable[[EventData], bool]] = {}
        self.enabled: bool = True
        self.debug: bool = debug

        if events:
            for ev in events:
                self.__listeners[ev] = []

        if self.debug:
            print(f"xEvents initialized with events: {events or []}")

    def channel(self, name: str) -> Channel:
        with self.__lock:
            if name not in self.__channels:
                self.__channels[name] = Channel(name, self)
            return self.__channels[name]

    def subscribe(
        self,
        event: Event,
        listener: Listener,
        once: bool = False,
        filter_fn: Callable[[EventData], bool] | None = None,
    ) -> None:
        with self.__lock:
            listeners = self.__listeners[event]
            if listener not in listeners:
                listeners.append(listener)
                key = _ListenerKey(event, id(listener))
                if once:
                    self.__once_listeners.add(key)
                if filter_fn:
                    self.__filters[key] = filter_fn
                if self.debug:
                    print(f"Listener {listener} added to event {event} (once={once})")

    def on(
        self,
        event: Event,
        listener: Listener,
        filter_fn: Callable[[EventData], bool] | None = None,
    ) -> None:
        self.subscribe(event, listener, filter_fn=filter_fn)

    def once(
        self,
        event: Event,
        listener: Listener,
        filter_fn: Callable[[EventData], bool] | None = None,
    ) -> None:
        self.subscribe(event, listener, once=True, filter_fn=filter_fn)

    def on_any(self, pattern: str, listener: Listener) -> None:
        with self.__lock:
            self.__wildcard_listeners.append((pattern, listener))
            if self.debug:
                print(f"Wildcard listener {listener} added for pattern {pattern!r}")

    def off_any(self, pattern: str, listener: Listener) -> None:
        with self.__lock:
            self.__wildcard_listeners = [
                (p, cb)
                for p, cb in self.__wildcard_listeners
                if not (p == pattern and cb is listener)
            ]

    def unsubscribe(self, event: Event, listener: Listener) -> None:
        with self.__lock:
            listeners = self.__listeners.get(event, [])
            if listener in listeners:
                listeners.remove(listener)
                key = _ListenerKey(event, id(listener))
                self.__once_listeners.discard(key)
                self.__filters.pop(key, None)

    def post(self, event: Event, data: EventData) -> None:
        with self.__lock:
            if not self.enabled:
                return

            record = EventRecord(
                event=event,
                data=data,
                channel=self._extract_channel(event),
            )
            self.__history.append(record)
            if len(self.__history) > self.__history_limit:
                self.__history = self.__history[-self.__history_limit:]

            listeners = list(self.__listeners.get(event, []))
            once_keys = set(self.__once_listeners)
            filters = dict(self.__filters)
            wildcards = list(self.__wildcard_listeners)

        to_remove: list[tuple[Event, Listener]] = []

        if self.debug:
            print(f"Event {event} posting to {len(listeners)} listeners")

        for cb in listeners:
            key = _ListenerKey(event, id(cb))
            filter_fn = filters.get(key)
            if filter_fn and not filter_fn(data):
                continue

            try:
                if asyncio.iscoroutinefunction(cb):
                    try:
                        loop = asyncio.get_running_loop()
                        loop.create_task(cb(data))
                    except RuntimeError:
                        asyncio.run(cb(data))
                else:
                    cb(data)
                if self.debug:
                    print(f"Event {event} delivered to {cb}")
            except KeyError as e:
                logger.error(
                    "KeyError in event '%s' listener=%r missing_key=%s payload_keys=%s",
                    event,
                    cb,
                    e.args[0],
                    list(data.keys()) if data else None,
                )
                logger.debug("Payload: %r", data, exc_info=True)
            except Exception:
                logger.exception(
                    "Unhandled error in event '%s' listener=%r payload=%r",
                    event,
                    cb,
                    data,
                )

            if key in once_keys:
                to_remove.append((event, cb))

        for pattern, cb in wildcards:
            if fnmatch(event, pattern):
                try:
                    if asyncio.iscoroutinefunction(cb):
                        try:
                            loop = asyncio.get_running_loop()
                            loop.create_task(cb(event, data))
                        except RuntimeError:
                            asyncio.run(cb(event, data))
                    else:
                        cb(event, data)
                except Exception:
                    logger.exception(
                        "Unhandled error in wildcard listener pattern=%r event=%r",
                        pattern,
                        event,
                    )

        if to_remove:
            with self.__lock:
                for ev, cb in to_remove:
                    key = _ListenerKey(ev, id(cb))
                    listeners_list = self.__listeners.get(ev, [])
                    if cb in listeners_list:
                        listeners_list.remove(cb)
                    self.__once_listeners.discard(key)
                    self.__filters.pop(key, None)

    def replay(
        self,
        event: Event,
        listener: Listener,
        limit: int | None = None,
    ) -> int:
        with self.__lock:
            matching = [r for r in self.__history if r.event == event]
            if limit:
                matching = matching[-limit:]
            records = list(matching)

        for record in records:
            try:
                listener(record.data)
            except Exception:
                logger.exception(
                    "Error during replay event=%r listener=%r",
                    event,
                    listener,
                )
        return len(records)

    def replay_channel(
        self,
        channel_name: str,
        listener: Listener,
        limit: int | None = None,
    ) -> int:
        with self.__lock:
            matching = [r for r in self.__history if r.channel == channel_name]
            if limit:
                matching = matching[-limit:]
            records = list(matching)

        for record in records:
            try:
                listener(record.data)
            except Exception:
                logger.exception(
                    "Error during channel replay channel=%r listener=%r",
                    channel_name,
                    listener,
                )
        return len(records)

    def history(
        self,
        event: Event | None = None,
        channel: str | None = None,
        limit: int | None = None,
    ) -> list[EventRecord]:
        with self.__lock:
            records = list(self.__history)
        if event:
            records = [r for r in records if r.event == event]
        if channel:
            records = [r for r in records if r.channel == channel]
        if limit:
            records = records[-limit:]
        return records

    def clear_history(self) -> None:
        with self.__lock:
            self.__history.clear()

    def events(self) -> list[Event]:
        with self.__lock:
            return list(self.__listeners.keys())

    def channels(self) -> list[str]:
        with self.__lock:
            return list(self.__channels.keys())

    def listener_count(self, event: Event) -> int:
        with self.__lock:
            return len(self.__listeners.get(event, []))

    def reset(self) -> None:
        with self.__lock:
            self.__listeners.clear()
            self.__once_listeners.clear()
            self.__wildcard_listeners.clear()
            self.__history.clear()
            self.__channels.clear()
            self.__filters.clear()

    def _extract_channel(self, event: Event) -> str | None:
        if CHANNEL_SEP in event:
            return event.split(CHANNEL_SEP, 1)[0]
        return None



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from xevents.constants import CHANNEL_SEP
from xevents.types import Event, EventData, Listener

if TYPE_CHECKING:
    from xevents.bus import xEvents


class Channel:
    def __init__(self, name: str, parent: xEvents):
        self.name = name
        self._parent = parent

    def subscribe(
        self,
        event: Event,
        listener: Listener,
        once: bool = False,
        filter_fn: Callable[[EventData], bool] | None = None,
    ) -> None:
        self._parent.subscribe(self._scoped(event), listener, once=once, filter_fn=filter_fn)

    def unsubscribe(self, event: Event, listener: Listener) -> None:
        self._parent.unsubscribe(self._scoped(event), listener)

    def post(self, event: Event, data: EventData) -> None:
        self._parent.post(self._scoped(event), data)

    def on(self, event: Event, listener: Listener) -> None:
        self.subscribe(event, listener)

    def once(self, event: Event, listener: Listener) -> None:
        self.subscribe(event, listener, once=True)

    def _scoped(self, event: Event) -> Event:
        return f"{self.name}{CHANNEL_SEP}{event}"

    def __repr__(self) -> str:
        return f"Channel({self.name!r})"
